@@ -109,7 +109,7 @@ public:
    :Item_str_ascii_func(thd, a) { }
   Item_str_ascii_checksum_func(THD *thd, Item *a, Item *b)
    :Item_str_ascii_func(thd, a, b) { }
-  bool eq(const Item *item, bool binary_cmp) const override
+  bool eq(const Item *item, bool binary_cmp) const
   {
     // Always use binary argument comparison: MD5('x') != MD5('X')
     return Item_func::eq(item, true);
@@ -129,7 +129,13 @@ public:
    :Item_str_func(thd, a) { }
   Item_str_binary_checksum_func(THD *thd, Item *a, Item *b)
    :Item_str_func(thd, a, b) { }
-  bool eq(const Item *item, bool binary_cmp) const override
+  Item_str_binary_checksum_func(THD *thd, Item *a, Item *b, Item *c)
+   :Item_str_func(thd, a, b, c) { }
+  Item_str_binary_checksum_func(THD *thd, Item *a, Item *b, Item *c, Item *d)
+   :Item_str_func(thd, a, b, c, d) { }
+  Item_str_binary_checksum_func(THD *thd, Item *a, Item *b, Item *c, Item *d, Item *e)
+   :Item_str_func(thd, a, b, c, d, e) { }
+  bool eq(const Item *item, bool binary_cmp) const
   {
     /*
       Always use binary argument comparison:
@@ -155,7 +161,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("md5") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_md5>(thd, this); }
 };
 
@@ -171,7 +177,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("sha") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_sha>(thd, this); }
 };
 
@@ -187,7 +193,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("sha2") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_sha2>(thd, this); }
 };
 
@@ -204,7 +210,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("to_base64") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_to_base64>(thd, this); }
 };
 
@@ -221,7 +227,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("from_base64") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_from_base64>(thd, this); }
 };
 
@@ -229,30 +235,46 @@ public:
 
 class Item_aes_crypt :public Item_str_binary_checksum_func
 {
-  enum { AES_KEY_LENGTH = 128 };
   void create_key(String *user_key, uchar* key);
-
-protected:
-  int what;
+  int parse_mode();
   String tmp_value;
+  const int what;
+  uint aes_key_length;
+  enum my_aes_mode aes_mode;
+
 public:
-  Item_aes_crypt(THD *thd, Item *a, Item *b)
-   :Item_str_binary_checksum_func(thd, a, b) {}
+  Item_aes_crypt(THD *thd, int what, Item *a, Item *b)
+   : Item_str_binary_checksum_func(thd, a, b), what(what) {}
+  Item_aes_crypt(THD *thd, int what, Item *a, Item *b, Item *c)
+   : Item_str_binary_checksum_func(thd, a, b, c), what(what) {}
+  Item_aes_crypt(THD *thd, int what, Item *a, Item *b, Item *c, Item *d)
+   : Item_str_binary_checksum_func(thd, a, b, c, d), what(what) {}
+  bool fix_fields(THD *thd, Item **ref) override;
   String *val_str(String *) override;
+  bool check_vcol_func_processor(void *arg) override
+  {
+    if (arg_count > 3)
+      return FALSE;
+    return mark_unsupported_function(func_name(), "()", arg, VCOL_SESSION_FUNC);
+  }
 };
 
 class Item_func_aes_encrypt :public Item_aes_crypt
 {
 public:
   Item_func_aes_encrypt(THD *thd, Item *a, Item *b)
-   :Item_aes_crypt(thd, a, b) {}
+   :Item_aes_crypt(thd, ENCRYPTION_FLAG_ENCRYPT, a, b) {}
+  Item_func_aes_encrypt(THD *thd, Item *a, Item *b, Item *c)
+   :Item_aes_crypt(thd, ENCRYPTION_FLAG_ENCRYPT, a, b, c) {}
+  Item_func_aes_encrypt(THD *thd, Item *a, Item *b, Item *c, Item *d)
+   :Item_aes_crypt(thd, ENCRYPTION_FLAG_ENCRYPT, a, b, c, d) {}
   bool fix_length_and_dec(THD *thd) override;
   LEX_CSTRING func_name_cstring() const override
   {
     static LEX_CSTRING name= {STRING_WITH_LEN("aes_encrypt") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_aes_encrypt>(thd, this); }
 };
 
@@ -260,17 +282,50 @@ class Item_func_aes_decrypt :public Item_aes_crypt
 {
 public:
   Item_func_aes_decrypt(THD *thd, Item *a, Item *b):
-    Item_aes_crypt(thd, a, b) {}
+    Item_aes_crypt(thd, ENCRYPTION_FLAG_DECRYPT, a, b) {}
+  Item_func_aes_decrypt(THD *thd, Item *a, Item *b, Item *c):
+    Item_aes_crypt(thd, ENCRYPTION_FLAG_DECRYPT, a, b, c) {}
+  Item_func_aes_decrypt(THD *thd, Item *a, Item *b, Item *c, Item *d):
+    Item_aes_crypt(thd, ENCRYPTION_FLAG_DECRYPT, a, b, c, d) {}
   bool fix_length_and_dec(THD *thd) override;
   LEX_CSTRING func_name_cstring() const override
   {
     static LEX_CSTRING name= {STRING_WITH_LEN("aes_decrypt") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_aes_decrypt>(thd, this); }
 };
 
+class Item_func_kdf :public Item_str_binary_checksum_func
+{
+  uint key_length;
+public:
+  Item_func_kdf(THD *thd, Item *a, Item *b)
+   : Item_str_binary_checksum_func(thd, a, b) {}
+  Item_func_kdf(THD *thd, Item *a, Item *b, Item *c)
+   : Item_str_binary_checksum_func(thd, a, b, c) {}
+  Item_func_kdf(THD *thd, Item *a, Item *b, Item *c, Item *d)
+   : Item_str_binary_checksum_func(thd, a, b, c, d) {}
+  Item_func_kdf(THD *thd, Item *a, Item *b, Item *c, Item *d, Item *e)
+   : Item_str_binary_checksum_func(thd, a, b, c, d, e) {}
+  bool fix_length_and_dec(THD *thd) override;
+  String *val_str(String *) override;
+  bool check_vcol_func_processor(void *arg) override
+  {
+    if (arg_count > 4)
+      return FALSE;
+    return mark_unsupported_function(func_name(), "()", arg, VCOL_SESSION_FUNC);
+  }
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("kdf") };
+    return name;
+  }
+  Item *get_copy(THD *thd) override
+  { return get_item_copy<Item_func_kdf>(thd, this); }
+};
+
 class Item_func_natural_sort_key : public Item_str_func
 {
 public:
@@ -283,7 +338,7 @@ public:
     return name;
   }
   bool fix_length_and_dec(THD *thd) override;
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   {
     return get_item_copy<Item_func_natural_sort_key>(thd, this);
   }
@@ -320,7 +375,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("concat") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_concat>(thd, this); }
 };
 
@@ -351,7 +406,7 @@ public:
       print_sql_mode_qualified_name(str, query_type);
     print_args_parenthesized(str, query_type);
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   {
     return get_item_copy<Item_func_concat_operator_oracle>(thd, this);
   }
@@ -376,7 +431,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("decode_histogram") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_decode_histogram>(thd, this); }
 };
 
@@ -393,7 +448,7 @@ public:
     return name;
   }
   table_map not_null_tables() const override { return 0; }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_concat_ws>(thd, this); }
 };
 
@@ -412,10 +467,9 @@ public:
   }
   bool check_vcol_func_processor(void *arg) override
   {
-    return mark_unsupported_function(func_name(), "()", arg,
-                                     VCOL_NON_DETERMINISTIC | VCOL_NEXTVAL);
+    return mark_unsupported_function(func_name(), "()", arg, VCOL_NON_DETERMINISTIC);
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   {
     return get_item_copy<Item_func_random_bytes>(thd, this);
   }
@@ -435,7 +489,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("reverse") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_reverse>(thd, this); }
 };
 
@@ -450,6 +504,7 @@ public:
     Item_str_func(thd, org, find, replace) {}
   String *val_str(String *to) override { return val_str_internal(to, false); };
   bool fix_length_and_dec(THD *thd) override;
+  String *val_str_internal(String *str, String *empty_string_for_null);
   const Schema *schema() const override { return &mariadb_schema; }
   void print(String *str, enum_query_type query_type) override
   {
@@ -461,7 +516,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("replace") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_replace>(thd, this); }
 };
 
@@ -486,7 +541,7 @@ public:
       print_sql_mode_qualified_name(str, query_type);
     print_args_parenthesized(str, query_type);
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_replace_oracle>(thd, this); }
 };
 
@@ -526,7 +581,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("regexp_replace") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override { return 0;}
+  Item *get_copy(THD *thd) override { return 0;}
 };
 
 
@@ -536,14 +591,14 @@ public:
   Item_func_regexp_replace_oracle(THD *thd, Item *a, Item *b, Item *c)
    :Item_func_regexp_replace(thd, a, b, c)
   {}
-  const Schema *schema() const override { return &oracle_schema_ref; }
-  bool fix_length_and_dec(THD *thd) override
+  const Schema *schema() const { return &oracle_schema_ref; }
+  bool fix_length_and_dec(THD *thd)
   {
     bool rc= Item_func_regexp_replace::fix_length_and_dec(thd);
     set_maybe_null(); // Empty result is converted to NULL
     return rc;
   }
-  String *val_str(String *str) override
+  String *val_str(String *str)
   {
     return val_str_internal(str, true);
   }
@@ -571,7 +626,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("regexp_substr") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override { return 0; }
+  Item *get_copy(THD *thd) override { return 0; }
 };
 
 
@@ -589,7 +644,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("insert") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_insert>(thd, this); }
 };
 
@@ -616,7 +671,7 @@ public:
     return name;
   }
   bool fix_length_and_dec(THD *thd) override;
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_lcase>(thd, this); }
 };
 
@@ -630,7 +685,7 @@ public:
     return name;
   }
   bool fix_length_and_dec(THD *thd) override;
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_ucase>(thd, this); }
 };
 
@@ -648,7 +703,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("left") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_left>(thd, this); }
 };
 
@@ -665,7 +720,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("right") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_right>(thd, this); }
 };
 
@@ -694,7 +749,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("substr") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_substr>(thd, this); }
 };
 
@@ -711,7 +766,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("sformat") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_sformat>(thd, this); }
 };
 
@@ -747,7 +802,7 @@ public:
       print_sql_mode_qualified_name(str, query_type);
     print_args_parenthesized(str, query_type);
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_substr_oracle>(thd, this); }
 };
 
@@ -764,7 +819,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("substring_index") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_substr_index>(thd, this); }
 
 };
@@ -807,7 +862,7 @@ public:
   }
   void print(String *str, enum_query_type query_type) override;
   virtual LEX_CSTRING mode_name() const { return { "both", 4}; }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_trim>(thd, this); }
 };
 
@@ -828,7 +883,7 @@ public:
     set_maybe_null();
     return res;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_trim_oracle>(thd, this); }
 };
 
@@ -851,7 +906,7 @@ public:
   }
   LEX_CSTRING mode_name() const override
   { return { STRING_WITH_LEN("leading") }; }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_ltrim>(thd, this); }
 };
 
@@ -872,7 +927,7 @@ public:
     set_maybe_null();
     return res;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_ltrim_oracle>(thd, this); }
 };
 
@@ -891,7 +946,7 @@ public:
   }
   LEX_CSTRING mode_name() const override
   { return { STRING_WITH_LEN("trailing") }; }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_rtrim>(thd, this); }
 };
 
@@ -912,7 +967,7 @@ public:
     set_maybe_null();
     return res;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_rtrim_oracle>(thd, this); }
 };
 
@@ -955,7 +1010,7 @@ public:
   }
   static char *alloc(THD *thd, const char *password, size_t pass_len,
                      enum PW_Alg al);
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_password>(thd, this); }
 };
 
@@ -976,7 +1031,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("des_encrypt") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_des_encrypt>(thd, this); }
 };
 
@@ -995,7 +1050,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("des_decrypt") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_des_decrypt>(thd, this); }
 };
 
@@ -1040,7 +1095,7 @@ public:
   {
     return FALSE;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_encrypt>(thd, this); }
 };
 
@@ -1064,7 +1119,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("encode") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_encode>(thd, this); }
 protected:
   virtual void crypto_transform(String *);
@@ -1089,7 +1144,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("decode") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_decode>(thd, this); }
 protected:
   void crypto_transform(String *) override;
@@ -1101,19 +1156,19 @@ class Item_func_sysconst :public Item_str_func
 public:
   Item_func_sysconst(THD *thd): Item_str_func(thd)
   { collation.set(system_charset_info,DERIVATION_SYSCONST); }
-  Item *safe_charset_converter(THD *thd, CHARSET_INFO *tocs) override;
+  Item *safe_charset_converter(THD *thd, CHARSET_INFO *tocs);
   /*
     Used to create correct Item name in new converted item in
     safe_charset_converter, return string representation of this function
     call
   */
   virtual const char *fully_qualified_func_name() const = 0;
-  bool check_vcol_func_processor(void *arg) override
+  bool check_vcol_func_processor(void *arg)
   {
     return mark_unsupported_function(fully_qualified_func_name(), arg,
                                      VCOL_SESSION_FUNC);
   }
-  bool const_item() const override;
+  bool const_item() const;
 };
 
 
@@ -1135,7 +1190,7 @@ public:
   }
   const char *fully_qualified_func_name() const override
   { return "database()"; }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_database>(thd, this); }
 };
 
@@ -1163,7 +1218,7 @@ public:
     base_flags&= ~item_base_t::MAYBE_NULL;
     return FALSE;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_sqlerrm>(thd, this); }
 };
 
@@ -1201,7 +1256,7 @@ public:
   {
     return save_str_value_in_field(field, &str_value);
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_user>(thd, this); }
 };
 
@@ -1263,7 +1318,7 @@ public:
     return mark_unsupported_function(fully_qualified_func_name(), arg,
                                      VCOL_SESSION_FUNC);
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_current_role>(thd, this); }
 };
 
@@ -1280,7 +1335,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("soundex") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_soundex>(thd, this); }
 };
 
@@ -1298,7 +1353,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("elt") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_elt>(thd, this); }
 };
 
@@ -1316,7 +1371,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("make_set") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_make_set>(thd, this); }
 };
 
@@ -1337,7 +1392,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("format") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_format>(thd, this); }
 };
 
@@ -1366,7 +1421,7 @@ public:
     return name;
   }
   void print(String *str, enum_query_type query_type) override;
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_char>(thd, this); }
 };
 
@@ -1386,7 +1441,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("chr") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_chr>(thd, this); }
 };
 
@@ -1403,7 +1458,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("repeat") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_repeat>(thd, this); }
 };
 
@@ -1419,7 +1474,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("space") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_space>(thd, this); }
 };
 
@@ -1440,7 +1495,7 @@ public:
   {
     return mark_unsupported_function(func_name(), "()", arg, VCOL_IMPOSSIBLE);
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_binlog_gtid_pos>(thd, this); }
 };
 
@@ -1482,7 +1537,7 @@ public:
     return name;
   }
   Sql_mode_dependency value_depends_on_sql_mode() const override;
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_rpad>(thd, this); }
 };
 
@@ -1516,7 +1571,7 @@ public:
       print_sql_mode_qualified_name(str, query_type);
     print_args_parenthesized(str, query_type);
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_rpad_oracle>(thd, this); }
 };
 
@@ -1542,7 +1597,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("lpad") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_lpad>(thd, this); }
 };
 
@@ -1576,7 +1631,7 @@ public:
       print_sql_mode_qualified_name(str, query_type);
     print_args_parenthesized(str, query_type);
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_lpad_oracle>(thd, this); }
 };
 
@@ -1599,7 +1654,7 @@ public:
     set_maybe_null();
     return FALSE;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_conv>(thd, this); }
 };
 
@@ -1633,23 +1688,13 @@ public:
   }
   bool fix_length_and_dec(THD *thd) override
   {
-    m_arg0_type_handler= args[0]->type_handler();
     collation.set(default_charset(), DERIVATION_COERCIBLE, MY_REPERTOIRE_ASCII);
     decimals=0;
-    /*
-      Reserve space for 16 characters for signed numeric data types:
-        hex(-1) -> 'FFFFFFFFFFFFFFFF'.
-      For unsigned numeric types, HEX() can create too large columns.
-      This should be eventually fixed to create minimum possible columns.
-    */
-    const Type_handler_numeric *tn=
-      dynamic_cast<const Type_handler_numeric*>(m_arg0_type_handler);
-    size_t char_length= (tn && !(tn->flags() & UNSIGNED_FLAG)) ?
-                        (size_t) 16 : (size_t) args[0]->max_length * 2;
-    fix_char_length(char_length);
+    fix_char_length(args[0]->max_length * 2);
+    m_arg0_type_handler= args[0]->type_handler();
     return FALSE;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_hex>(thd, this); }
 };
 
@@ -1675,7 +1720,7 @@ public:
     max_length=(1+args[0]->max_length)/2;
     return FALSE;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_unhex>(thd, this); }
 };
 
@@ -1714,7 +1759,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("like_range_min") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_like_range_min>(thd, this); }
 };
 
@@ -1729,7 +1774,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("like_range_max") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_like_range_max>(thd, this); }
 };
 #endif
@@ -1761,7 +1806,7 @@ public:
     return name;
   }
   bool need_parentheses_in_default() override { return true; }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_binary>(thd, this); }
 };
 
@@ -1788,7 +1833,7 @@ public:
   {
     return mark_unsupported_function(func_name(), "()", arg, VCOL_IMPOSSIBLE);
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_load_file>(thd, this); }
 };
 
@@ -1809,7 +1854,7 @@ class Item_func_export_set: public Item_str_func
     static LEX_CSTRING name= {STRING_WITH_LEN("export_set") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_export_set>(thd, this); }
 };
 
@@ -1836,7 +1881,7 @@ public:
     max_length= (uint32) MY_MIN(max_result_length, MAX_BLOB_WIDTH);
     return FALSE;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_quote>(thd, this); }
 };
 
@@ -1932,7 +1977,7 @@ public:
     return name;
   }
   void print(String *str, enum_query_type query_type) override;
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_conv_charset>(thd, this); }
   int save_in_field(Field*, bool) override;
 };
@@ -1961,7 +2006,7 @@ public:
     return args[0]->field_for_view_update();
   }
   bool need_parentheses_in_default() override { return true; }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_set_collation>(thd, this); }
 };
 
@@ -1987,8 +2032,6 @@ public:
 
 class Item_func_charset :public Item_func_expr_str_metadata
 {
-  LEX_CSTRING m_cached_charset_info;
-
 public:
   Item_func_charset(THD *thd, Item *a)
     :Item_func_expr_str_metadata(thd, a) { }
@@ -1998,25 +2041,8 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("charset") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_charset>(thd, this); }
-  table_map used_tables() const override { return 0; }
-  bool fix_length_and_dec(THD* thd) override
-  {
-    if (Item_func_expr_str_metadata::fix_length_and_dec(thd))
-      return true;
-    /*
-      Since this is a const item which doesn't use tables (see used_tables()),
-      we don't want to access the function arguments during execution.
-      That's why we store the charset here during the preparation phase
-      and only return it later at the execution phase
-    */
-    DBUG_ASSERT(args[0]->fixed());
-    m_cached_charset_info.str= args[0]->charset_for_protocol()->cs_name.str;
-    m_cached_charset_info.length=
-        args[0]->charset_for_protocol()->cs_name.length;
-    return false;
-  }
 };
 
 
@@ -2031,7 +2057,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("collation") };
     return name;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_collation>(thd, this); }
 };
 
@@ -2071,7 +2097,7 @@ public:
     override
   { return this; }
   void print(String *str, enum_query_type query_type) override;
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_weight_string>(thd, this); }
 };
 
@@ -2102,7 +2128,7 @@ public:
   }
   bool fix_length_and_dec(THD *thd) override { max_length=10; return FALSE; }
   longlong val_int() override;
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_crc32>(thd, this); }
 };
 
@@ -2123,7 +2149,7 @@ public:
     set_maybe_null();
     return FALSE; }
   longlong val_int() override;
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_uncompressed_length>(thd, this); }
 };
 
@@ -2150,7 +2176,7 @@ public:
     return name;
   }
   String *val_str(String *) override ZLIB_DEPENDED_FUNCTION
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_compress>(thd, this); }
 };
 
@@ -2172,7 +2198,7 @@ public:
     return name;
   }
   String *val_str(String *) override ZLIB_DEPENDED_FUNCTION
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_uncompress>(thd, this); }
 };
 
@@ -2199,7 +2225,7 @@ public:
   String *val_str(String *) override;
   void print(String *str, enum_query_type query_type) override;
   enum Functype functype() const override { return DYNCOL_FUNC; }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_dyncol_create>(thd, this); }
 };
 
@@ -2217,7 +2243,7 @@ public:
   }
   String *val_str(String *) override;
   void print(String *str, enum_query_type query_type) override;
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_dyncol_add>(thd, this); }
 };
 
@@ -2239,7 +2265,7 @@ public:
     decimals= 0;
     return FALSE;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_dyncol_json>(thd, this); }
 };
 
@@ -2289,7 +2315,7 @@ public:
   bool get_dyn_value(THD *thd, DYNAMIC_COLUMN_VALUE *val, String *tmp);
   bool get_date(THD *thd, MYSQL_TIME *ltime, date_mode_t fuzzydate) override;
   void print(String *str, enum_query_type query_type) override;
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_dyncol_get>(thd, this); }
 };
 
@@ -2311,7 +2337,7 @@ public:
     return name;
   }
   String *val_str(String *) override;
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_dyncol_list>(thd, this); }
 };
 
@@ -2337,9 +2363,36 @@ public:
     return name;
   }
   bool fix_length_and_dec(THD *thd) override;
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_temptable_rowid>(thd, this); }
 };
+
+class Item_func_format_pico_time : public Item_str_ascii_func
+{
+  /* Format is 'AAAA.BB UUU' = 11 characters or 'AAA ps' = 6 characters. */
+  char m_value_buffer[12];
+  String m_value;
+
+public:
+  Item_func_format_pico_time(THD *thd, Item *a): Item_str_ascii_func(thd, a) {}
+  String *val_str_ascii(String *) override;
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("format_pico_time")};
+    return name;
+  }
+  bool fix_length_and_dec(THD *thd) override
+  {
+    m_value.set(m_value_buffer, sizeof(m_value_buffer), default_charset());
+    fix_length_and_charset(sizeof(m_value_buffer), default_charset());
+    return false;
+  }
+  Item *get_copy(THD *thd) override
+  {
+    return get_item_copy<Item_func_format_pico_time>(thd, this);
+  }
+};
+
 #ifdef WITH_WSREP
 
 #include "wsrep_api.h"
@@ -2361,7 +2414,7 @@ public:
     set_maybe_null();
     return FALSE;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_wsrep_last_written_gtid>(thd, this); }
 };
 
@@ -2382,7 +2435,7 @@ public:
     set_maybe_null();
     return FALSE;
   }
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_wsrep_last_seen_gtid>(thd, this); }
 };
 
@@ -2400,7 +2453,7 @@ public:
     return name;
   }
   longlong val_int() override;
-  Item *do_get_copy(THD *thd) const override
+  Item *get_copy(THD *thd) override
   { return get_item_copy<Item_func_wsrep_sync_wait_upto>(thd, this); }
 };
 #endif /* WITH_WSREP */
