@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 1995, 2021, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2014, 2022, MariaDB Corporation.
+Copyright (c) 2014, 2023, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -45,10 +45,9 @@ Created 10/25/1995 Heikki Tuuri
 #include "srv0start.h"
 #include "trx0purge.h"
 #include "buf0lru.h"
-#include "ibuf0ibuf.h"
 #include "buf0flu.h"
 #include "log.h"
-#ifdef UNIV_LINUX
+#ifdef __linux__
 # include <sys/types.h>
 # include <sys/sysmacros.h>
 # include <dirent.h>
@@ -97,7 +96,11 @@ bool fil_space_t::try_to_close(bool print_info)
     if (!node->is_open())
       continue;
 
-    if (const auto n= space.set_closing())
+    const auto n= space.set_closing();
+    if (n & STOPPING)
+      /* Let fil_space_t::drop() in another thread handle this. */
+      continue;
+    if (n & (PENDING | NEEDS_FSYNC))
     {
       if (!print_info)
         continue;
@@ -119,6 +122,9 @@ bool fil_space_t::try_to_close(bool print_info)
     }
 
     node->close();
+
+    fil_system.move_closed_last_to_space_list(node->space);
+
     return true;
   }
 
@@ -286,6 +292,8 @@ fil_node_t* fil_space_t::add(const char* name, pfs_os_file_t handle,
 			     uint32_t size, bool is_raw, bool atomic_write,
 			     uint32_t max_pages)
 {
+	mysql_mutex_assert_owner(&fil_system.mutex);
+
 	fil_node_t*	node;
 
 	ut_ad(name != NULL);
@@ -310,7 +318,6 @@ fil_node_t* fil_space_t::add(const char* name, pfs_os_file_t handle,
 
 	node->atomic_write = atomic_write;
 
-	mysql_mutex_lock(&fil_system.mutex);
 	this->size += size;
 	UT_LIST_ADD_LAST(chain, node);
 	if (node->is_open()) {
@@ -321,11 +328,11 @@ fil_node_t* fil_space_t::add(const char* name, pfs_os_file_t handle,
 			release();
 		}
 	}
-	mysql_mutex_unlock(&fil_system.mutex);
 
 	return node;
 }
 
+__attribute__((warn_unused_result, nonnull))
 /** Open a tablespace file.
 @param node  data file
 @return whether the file was successfully opened */
@@ -334,8 +341,9 @@ static bool fil_node_open_file_low(fil_node_t *node)
   ut_ad(!node->is_open());
   ut_ad(node->space->is_closing());
   mysql_mutex_assert_owner(&fil_system.mutex);
-  ulint type;
   static_assert(((UNIV_ZIP_SIZE_MIN >> 1) << 3) == 4096, "compatibility");
+#if defined _WIN32 || defined HAVE_FCNTL_DIRECT
+  ulint type;
   switch (FSP_FLAGS_GET_ZIP_SSIZE(node->space->flags)) {
   case 1:
   case 2:
@@ -344,6 +352,9 @@ static bool fil_node_open_file_low(fil_node_t *node)
   default:
     type= OS_DATA_FILE;
   }
+#else
+  constexpr auto type= OS_DATA_FILE;
+#endif
 
   for (;;)
   {
@@ -354,8 +365,20 @@ static bool fil_node_open_file_low(fil_node_t *node)
                                  : OS_FILE_OPEN | OS_FILE_ON_ERROR_NO_EXIT,
                                  OS_FILE_AIO, type,
                                  srv_read_only_mode, &success);
-    if (success)
+
+    if (success && node->is_open())
+    {
+#ifndef _WIN32
+      if (!node->space->id && !srv_read_only_mode && my_disable_locking &&
+          os_file_lock(node->handle, node->name))
+      {
+        os_file_close(node->handle);
+        node->handle= OS_FILE_CLOSED;
+        return false;
+      }
+#endif
       break;
+    }
 
     /* The following call prints an error message */
     if (os_file_get_last_error(true) == EMFILE + 100 &&
@@ -393,13 +416,7 @@ static bool fil_node_open_file_low(fil_node_t *node)
 
   ut_ad(node->is_open());
 
-  if (UNIV_LIKELY(!fil_system.freeze_space_list))
-  {
-    /* Move the file last in fil_system.space_list, so that
-    fil_space_t::try_to_close() should close it as a last resort. */
-    fil_system.space_list.erase(space_list_t::iterator(node->space));
-    fil_system.space_list.push_back(*node->space);
-  }
+  fil_system.move_opened_last_to_space_list(node->space);
 
   fil_system.n_open++;
   return true;
@@ -445,7 +462,9 @@ static bool fil_node_open_file(fil_node_t *node)
     }
   }
 
-  return fil_node_open_file_low(node);
+  /* The node can be opened beween releasing and acquiring fil_system.mutex
+  in the above code */
+  return node->is_open() || fil_node_open_file_low(node);
 }
 
 /** Close the file handle. */
@@ -493,12 +512,15 @@ void fil_space_t::flush_low()
                                             std::memory_order_relaxed))
   {
     ut_ad(n & PENDING);
-    if (n & STOPPING)
+    if (n & STOPPING_WRITES)
       return;
     if (n & NEEDS_FSYNC)
       break;
   }
 
+  if (fil_system.is_write_through())
+    goto skip_flush;
+
   fil_n_pending_tablespace_flushes++;
   for (fil_node_t *node= UT_LIST_GET_FIRST(chain); node;
        node= UT_LIST_GET_NEXT(chain, node))
@@ -523,8 +545,9 @@ void fil_space_t::flush_low()
     mysql_mutex_unlock(&fil_system.mutex);
   }
 
-  clear_flush();
   fil_n_pending_tablespace_flushes--;
+skip_flush:
+  clear_flush();
 }
 
 /** Try to extend a tablespace.
@@ -545,7 +568,7 @@ fil_space_extend_must_retry(
 	ut_ad(UT_LIST_GET_LAST(space->chain) == node);
 	ut_ad(size >= FIL_IBD_FILE_INITIAL_SIZE);
 	ut_ad(node->space == space);
-	ut_ad(space->referenced() || space->is_being_truncated);
+	ut_ad(space->referenced());
 
 	*success = space->size >= size;
 
@@ -634,8 +657,7 @@ fil_space_extend_must_retry(
 	default:
 		ut_ad(space->purpose == FIL_TYPE_TABLESPACE
 		      || space->purpose == FIL_TYPE_IMPORT);
-		if (space->purpose == FIL_TYPE_TABLESPACE
-		    && !space->is_being_truncated) {
+		if (space->purpose == FIL_TYPE_TABLESPACE) {
 			goto do_flush;
 		}
 		break;
@@ -648,6 +670,19 @@ fil_space_extend_must_retry(
 	return false;
 }
 
+bool recv_sys_t::check_sys_truncate()
+{
+  mysql_mutex_assert_owner(&fil_system.mutex);
+  if (!truncated_sys_space.lsn)
+    return false;
+  if (fil_system.sys_space->size <= fil_system.sys_space->recv_size)
+  {
+    truncated_sys_space={0,0};
+    return false;
+  }
+  return true;
+}
+
 /** @return whether the file is usable for io() */
 ATTRIBUTE_COLD bool fil_space_t::prepare_acquired()
 {
@@ -661,8 +696,11 @@ ATTRIBUTE_COLD bool fil_space_t::prepare_acquired()
 
   if (!is_open)
     release();
+  else if (node->deferred);
   else if (auto desired_size= recv_size)
   {
+    if (id == TRX_SYS_SPACE && recv_sys.check_sys_truncate())
+      goto clear;
     bool success;
     while (fil_space_extend_must_retry(this, node, desired_size, &success))
       mysql_mutex_lock(&fil_system.mutex);
@@ -719,12 +757,10 @@ bool fil_space_extend(fil_space_t *space, uint32_t size)
   bool success= false;
   const bool acquired= space->acquire();
   mysql_mutex_lock(&fil_system.mutex);
-  if (acquired || space->is_being_truncated)
-  {
+  if (acquired)
     while (fil_space_extend_must_retry(space, UT_LIST_GET_LAST(space->chain),
                                        size, &success))
       mysql_mutex_lock(&fil_system.mutex);
-  }
   mysql_mutex_unlock(&fil_system.mutex);
   if (acquired)
     space->release();
@@ -752,7 +788,6 @@ inline pfs_os_file_t fil_node_t::close_to_free(bool detach_handle)
   {
     if (space->is_in_unflushed_spaces)
     {
-      ut_ad(srv_file_flush_method != SRV_O_DIRECT_NO_FSYNC);
       space->is_in_unflushed_spaces= false;
       fil_system.unflushed_spaces.remove(*space);
     }
@@ -785,7 +820,6 @@ pfs_os_file_t fil_system_t::detach(fil_space_t *space, bool detach_handle)
 
   if (space->is_in_unflushed_spaces)
   {
-    ut_ad(srv_file_flush_method != SRV_O_DIRECT_NO_FSYNC);
     space->is_in_unflushed_spaces= false;
     unflushed_spaces.remove(*space);
   }
@@ -795,7 +829,26 @@ pfs_os_file_t fil_system_t::detach(fil_space_t *space, bool detach_handle)
     space->is_in_default_encrypt= false;
     default_encrypt_tables.remove(*space);
   }
-  space_list.erase(space_list_t::iterator(space));
+
+  {
+    space_list_t::iterator s= space_list_t::iterator(space);
+    if (space_list_last_opened == space)
+    {
+      if (s == space_list.begin())
+      {
+        ut_ad(srv_operation > SRV_OPERATION_EXPORT_RESTORED ||
+              srv_shutdown_state > SRV_SHUTDOWN_NONE);
+        space_list_last_opened= nullptr;
+      }
+      else
+      {
+        space_list_t::iterator prev= s;
+        space_list_last_opened= &*--prev;
+      }
+    }
+    space_list.erase(s);
+  }
+
   if (space == sys_space)
     sys_space= nullptr;
   else if (space == temp_space)
@@ -883,18 +936,22 @@ bool fil_space_free(uint32_t id, bool x_latched)
 		}
 
 		if (!recv_recovery_is_on()) {
-			mysql_mutex_lock(&log_sys.mutex);
-		}
+			log_sys.latch.wr_lock(SRW_LOCK_CALL);
 
-		mysql_mutex_assert_owner(&log_sys.mutex);
+			if (space->max_lsn) {
+				ut_d(space->max_lsn = 0);
+				fil_system.named_spaces.remove(*space);
+			}
 
-		if (space->max_lsn != 0) {
-			ut_d(space->max_lsn = 0);
-			fil_system.named_spaces.remove(*space);
-		}
-
-		if (!recv_recovery_is_on()) {
-			mysql_mutex_unlock(&log_sys.mutex);
+			log_sys.latch.wr_unlock();
+		} else {
+#ifndef SUX_LOCK_GENERIC
+			ut_ad(log_sys.latch.is_write_locked());
+#endif
+			if (space->max_lsn) {
+				ut_d(space->max_lsn = 0);
+				fil_system.named_spaces.remove(*space);
+			}
 		}
 
 		fil_space_free_low(space);
@@ -910,15 +967,18 @@ bool fil_space_free(uint32_t id, bool x_latched)
 @param purpose    tablespace purpose
 @param crypt_data encryption information
 @param mode       encryption mode
+@param opened     true if space files are opened
 @return pointer to created tablespace, to be filled in with add()
 @retval nullptr on failure (such as when the same tablespace exists) */
 fil_space_t *fil_space_t::create(uint32_t id, uint32_t flags,
                                  fil_type_t purpose,
 				 fil_space_crypt_t *crypt_data,
-				 fil_encryption_t mode)
+				 fil_encryption_t mode,
+				 bool opened)
 {
 	fil_space_t*	space;
 
+	mysql_mutex_assert_owner(&fil_system.mutex);
 	ut_ad(fil_system.is_initialised());
 	ut_ad(fil_space_t::is_valid_flags(flags & ~FSP_FLAGS_MEM_MASK, id));
 	ut_ad(srv_page_size == UNIV_PAGE_SIZE_ORIG || flags != 0);
@@ -951,8 +1011,6 @@ fil_space_t *fil_space_t::create(uint32_t id, uint32_t flags,
 
 	space->latch.SRW_LOCK_INIT(fil_space_latch_key);
 
-	mysql_mutex_lock(&fil_system.mutex);
-
 	if (const fil_space_t *old_space = fil_space_get_by_id(id)) {
 		ib::error() << "Trying to add tablespace with id " << id
 			    << " to the cache, but tablespace '"
@@ -960,7 +1018,6 @@ fil_space_t *fil_space_t::create(uint32_t id, uint32_t flags,
 				? old_space->chain.start->name
 				: "")
 			    << "' already exists in the cache!";
-		mysql_mutex_unlock(&fil_system.mutex);
 		space->~fil_space_t();
 		ut_free(space);
 		return(NULL);
@@ -968,7 +1025,10 @@ fil_space_t *fil_space_t::create(uint32_t id, uint32_t flags,
 
 	HASH_INSERT(fil_space_t, hash, &fil_system.spaces, id, space);
 
-	fil_system.space_list.push_back(*space);
+	if (opened)
+	  fil_system.add_opened_last_to_space_list(space);
+	else
+          fil_system.space_list.push_back(*space);
 
 	switch (id) {
 	case 0:
@@ -1004,12 +1064,12 @@ fil_space_t *fil_space_t::create(uint32_t id, uint32_t flags,
 	if (rotate) {
 		fil_system.default_encrypt_tables.push_back(*space);
 		space->is_in_default_encrypt = true;
-	}
 
-	mysql_mutex_unlock(&fil_system.mutex);
-
-	if (rotate && srv_n_fil_crypt_threads_started) {
-		fil_crypt_threads_signal();
+		if (srv_n_fil_crypt_threads_started) {
+			mysql_mutex_unlock(&fil_system.mutex);
+			fil_crypt_threads_signal();
+			mysql_mutex_lock(&fil_system.mutex);
+		}
 	}
 
 	return(space);
@@ -1203,7 +1263,7 @@ void fil_system_t::create(ulint hash_size)
 	spaces.create(hash_size);
 
 	fil_space_crypt_init();
-#ifdef UNIV_LINUX
+#ifdef __linux__
 	ssd.clear();
 	char fn[sizeof(dirent::d_name)
 		+ sizeof "/sys/block/" "/queue/rotational"];
@@ -1283,10 +1343,19 @@ void fil_system_t::close()
 
   ut_ad(!spaces.array);
 
-#ifdef UNIV_LINUX
+#ifdef __linux__
   ssd.clear();
   ssd.shrink_to_fit();
-#endif /* UNIV_LINUX */
+#endif /* __linux__ */
+}
+
+void fil_system_t::add_opened_last_to_space_list(fil_space_t *space)
+{
+  if (UNIV_LIKELY(space_list_last_opened != nullptr))
+    space_list.insert(++space_list_t::iterator(space_list_last_opened), *space);
+  else
+    space_list.push_front(*space);
+  space_list_last_opened= space;
 }
 
 /** Extend all open data files to the recovered size */
@@ -1315,6 +1384,120 @@ ATTRIBUTE_COLD void fil_system_t::extend_to_recv_size()
   mysql_mutex_unlock(&mutex);
 }
 
+ATTRIBUTE_COLD void fil_space_t::reopen_all()
+{
+  mysql_mutex_assert_owner(&fil_system.mutex);
+  fil_system.freeze_space_list++;
+
+  for (fil_space_t &space : fil_system.space_list)
+  {
+    for (fil_node_t *node= UT_LIST_GET_FIRST(space.chain); node;
+         node= UT_LIST_GET_NEXT(chain, node))
+      if (node->is_open())
+        goto need_to_close;
+    continue;
+
+  need_to_close:
+    uint32_t p= space.n_pending.fetch_or(CLOSING, std::memory_order_acquire);
+    if (p & (STOPPING | CLOSING))
+      continue;
+
+    for (fil_node_t *node= UT_LIST_GET_FIRST(space.chain); node;
+         node= UT_LIST_GET_NEXT(chain, node))
+    {
+      if (!node->is_open())
+        continue;
+
+      ulint type= OS_DATA_FILE;
+
+      switch (FSP_FLAGS_GET_ZIP_SSIZE(space.flags)) {
+      case 1: case 2:
+        type= OS_DATA_FILE_NO_O_DIRECT;
+      }
+
+      for (ulint count= 10000; count--;)
+      {
+        p= space.pending();
+
+        if (!(p & CLOSING) || (p & STOPPING))
+          break;
+
+        if (!(p & PENDING) && !node->being_extended)
+        {
+          space.reacquire();
+          mysql_mutex_unlock(&fil_system.mutex);
+          /* Unconditionally flush the file, because
+          fil_system.write_through was updated prematurely,
+          potentially causing some flushes to be lost. */
+          os_file_flush(node->handle);
+          mysql_mutex_lock(&fil_system.mutex);
+          p= space.n_pending.fetch_sub(1, std::memory_order_relaxed) - 1;
+
+          if (!(p & CLOSING) || (p & STOPPING))
+            break;
+
+          if (!(p & PENDING) && !node->being_extended)
+          {
+            ut_a(os_file_close(node->handle));
+            bool success;
+            node->handle= os_file_create(innodb_data_file_key, node->name,
+                                         node->is_raw_disk
+                                         ? OS_FILE_OPEN_RAW : OS_FILE_OPEN,
+                                         OS_FILE_AIO, type,
+                                         srv_read_only_mode, &success);
+            ut_a(success);
+            goto next_file;
+          }
+        }
+
+        space.reacquire();
+        mysql_mutex_unlock(&fil_system.mutex);
+        std::this_thread::sleep_for(std::chrono::microseconds(100));
+        mysql_mutex_lock(&fil_system.mutex);
+        space.release();
+
+        if (!node->is_open())
+          goto next_file;
+      }
+
+      if (!(p & CLOSING) || (p & STOPPING))
+      next_file:
+        continue;
+
+      sql_print_error("InnoDB: Failed to reopen file '%s' due to " UINT32PF
+                      " operations", node->name, p & PENDING);
+    }
+  }
+
+  fil_system.freeze_space_list--;
+}
+
+void fil_system_t::set_write_through(bool write_through)
+{
+  mysql_mutex_lock(&mutex);
+
+  if (write_through != is_write_through())
+  {
+    this->write_through= write_through;
+    fil_space_t::reopen_all();
+  }
+
+  mysql_mutex_unlock(&mutex);
+}
+
+void fil_system_t::set_buffered(bool buffered)
+{
+  mysql_mutex_lock(&mutex);
+
+  if (buffered != is_buffered())
+  {
+    this->buffered= buffered;
+    fil_space_t::reopen_all();
+  }
+
+  mysql_mutex_unlock(&mutex);
+}
+
 /** Close all tablespace files at shutdown */
 void fil_space_t::close_all()
 {
@@ -1335,16 +1518,16 @@ void fil_space_t::close_all()
     for (fil_node_t *node= UT_LIST_GET_FIRST(space.chain); node != NULL;
          node= UT_LIST_GET_NEXT(chain, node))
     {
-
       if (!node->is_open())
-      {
       next:
         continue;
-      }
 
       for (ulint count= 10000; count--;)
       {
-        if (!space.set_closing())
+        const auto n= space.set_closing();
+        if (n & STOPPING)
+          goto next;
+        if (!(n & (PENDING | NEEDS_FSYNC)))
         {
           node->close();
           goto next;
@@ -1356,8 +1539,8 @@ void fil_space_t::close_all()
           goto next;
       }
 
-      ib::error() << "File '" << node->name << "' has " << space.referenced()
-                  << " operations";
+      sql_print_error("InnoDB: File '%s' has " UINT32PF " operations",
+                      node->name, space.referenced());
     }
 
     fil_system.detach(&space);
@@ -1389,46 +1572,6 @@ void fil_set_max_space_id_if_bigger(uint32_t max_id)
 	mysql_mutex_unlock(&fil_system.mutex);
 }
 
-/** Write the flushed LSN to the page header of the first page in the
-system tablespace.
-@param[in]	lsn	flushed LSN
-@return DB_SUCCESS or error number */
-dberr_t
-fil_write_flushed_lsn(
-	lsn_t	lsn)
-{
-	byte*	buf;
-	ut_ad(!srv_read_only_mode);
-
-	if (!fil_system.sys_space->acquire()) {
-		return DB_ERROR;
-	}
-
-	buf = static_cast<byte*>(aligned_malloc(srv_page_size, srv_page_size));
-
-	auto fio = fil_system.sys_space->io(IORequestRead, 0, srv_page_size,
-					    buf);
-
-	if (fio.err == DB_SUCCESS) {
-		mach_write_to_8(buf + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION,
-				lsn);
-
-		uint32_t fsp_flags = mach_read_from_4(
-			buf + FSP_HEADER_OFFSET + FSP_SPACE_FLAGS);
-
-		if (fil_space_t::full_crc32(fsp_flags)) {
-			buf_flush_assign_full_crc32_checksum(buf);
-		}
-
-		fio = fil_system.sys_space->io(IORequestWrite,
-					       0, srv_page_size, buf);
-		fil_flush_file_spaces();
-	}
-
-	aligned_free(buf);
-	return fio.err;
-}
-
 /** Acquire a tablespace reference.
 @param id      tablespace identifier
 @return tablespace
@@ -1464,7 +1607,7 @@ inline void mtr_t::log_file_op(mfile_type_t type, uint32_t space_id,
   ut_ad(strchr(path, '/'));
   ut_ad(!strcmp(&path[strlen(path) - strlen(DOT_IBD)], DOT_IBD));
 
-  flag_modified();
+  m_modifications= true;
   if (!is_logged())
     return;
   m_last= nullptr;
@@ -1501,7 +1644,7 @@ inline void mtr_t::log_file_op(mfile_type_t type, uint32_t space_id,
   {
     ut_ad(strchr(new_path, '/'));
     m_log.push(reinterpret_cast<const byte*>(path), uint32_t(len + 1));
-    m_log.push(reinterpret_cast<const byte*>(new_path), uint32_t(new_len));
+    m_log.push(reinterpret_cast<const byte*>(new_path), uint32_t(new_len - 1));
   }
   else
     m_log.push(reinterpret_cast<const byte*>(path), uint32_t(len));
@@ -1518,7 +1661,7 @@ static void fil_name_write(uint32_t space_id, const char *name,
   mtr->log_file_op(FILE_MODIFY, space_id, name);
 }
 
-fil_space_t *fil_space_t::check_pending_operations(uint32_t id)
+fil_space_t *fil_space_t::drop(uint32_t id, pfs_os_file_t *detached_handle)
 {
   ut_a(!is_system_tablespace(id));
   mysql_mutex_lock(&fil_system.mutex);
@@ -1532,7 +1675,6 @@ fil_space_t *fil_space_t::check_pending_operations(uint32_t id)
 
   if (space->pending() & STOPPING)
   {
-being_deleted:
     /* A thread executing DDL and another thread executing purge may
     be executing fil_delete_tablespace() concurrently for the same
     tablespace. Wait for the other thread to complete the operation. */
@@ -1551,34 +1693,83 @@ being_deleted:
       mysql_mutex_lock(&fil_system.mutex);
     }
   }
-  else
-  {
-    if (space->crypt_data)
-    {
-      space->reacquire();
-      mysql_mutex_unlock(&fil_system.mutex);
-      fil_space_crypt_close_tablespace(space);
-      mysql_mutex_lock(&fil_system.mutex);
-      space->release();
-    }
-    if (space->set_stopping_check())
-      goto being_deleted;
-  }
 
+  /* We must be the first one to set either STOPPING flag on the .ibd file,
+  because the flags are only being set here, within a critical section of
+  fil_system.mutex. */
+  unsigned pending;
+  ut_d(pending=)
+    space->n_pending.fetch_add(STOPPING_READS + 1, std::memory_order_relaxed);
+  ut_ad(!(pending & STOPPING));
   mysql_mutex_unlock(&fil_system.mutex);
 
-  for (ulint count= 0;; count++)
+  if (space->crypt_data)
+    fil_space_crypt_close_tablespace(space);
+
+  if (space->purpose == FIL_TYPE_TABLESPACE)
   {
-    const unsigned pending= space->referenced();
-    if (!pending)
-      return space;
-    /* Issue a warning every 10.24 seconds, starting after 2.56 seconds */
-    if ((count & 511) == 128)
-      sql_print_warning("InnoDB: Trying to delete tablespace '%s' "
-                        "but there are %u pending operations",
-                        space->chain.start->name, id);
-    std::this_thread::sleep_for(std::chrono::milliseconds(20));
+    if (id >= srv_undo_space_id_start &&
+        id < srv_undo_space_id_start + srv_undo_tablespaces_open)
+    {
+      os_file_delete(innodb_data_file_key, space->chain.start->name);
+      goto deleted;
+    }
+
+    /* Before deleting the file, persistently write a log record. */
+    mtr_t mtr;
+    mtr.start();
+    mtr.log_file_op(FILE_DELETE, id, space->chain.start->name);
+    mtr.commit_file(*space, nullptr);
+
+    if (FSP_FLAGS_HAS_DATA_DIR(space->flags))
+      RemoteDatafile::delete_link_file(space->name());
+
+    os_file_delete(innodb_data_file_key, space->chain.start->name);
   }
+  else
+    ut_ad(space->purpose == FIL_TYPE_IMPORT);
+
+  if (char *cfg_name= fil_make_filepath(space->chain.start->name,
+                                        fil_space_t::name_type{}, CFG, false))
+  {
+    os_file_delete_if_exists(innodb_data_file_key, cfg_name, nullptr);
+    ut_free(cfg_name);
+  }
+
+ deleted:
+  mysql_mutex_lock(&fil_system.mutex);
+  ut_ad(space == fil_space_get_by_id(id));
+  pending=
+    space->n_pending.fetch_add(STOPPING_WRITES - 1, std::memory_order_relaxed);
+  ut_ad((pending & STOPPING) == STOPPING_READS);
+  ut_ad(pending & PENDING);
+  pending&= PENDING;
+  if (--pending)
+  {
+    for (ulint count= 0;; count++)
+    {
+      ut_ad(space == fil_space_get_by_id(id));
+      pending= space->n_pending.load(std::memory_order_relaxed) & PENDING;
+      if (!pending)
+        break;
+      mysql_mutex_unlock(&fil_system.mutex);
+      /* Issue a warning every 10.24 seconds, starting after 2.56 seconds */
+      if ((count & 511) == 128)
+        sql_print_warning("InnoDB: Trying to delete tablespace '%s' "
+                          "but there are %u pending operations",
+                          space->chain.start->name, pending);
+      std::this_thread::sleep_for(std::chrono::milliseconds(20));
+      mysql_mutex_lock(&fil_system.mutex);
+    }
+  }
+
+  pfs_os_file_t handle= fil_system.detach(space, true);
+  mysql_mutex_unlock(&fil_system.mutex);
+  if (detached_handle)
+    *detached_handle = handle;
+  else
+    os_file_close(handle);
+  return space;
 }
 
 /** Close a single-table tablespace on failed IMPORT TABLESPACE.
@@ -1587,37 +1778,29 @@ Free all pages used by the tablespace. */
 void fil_close_tablespace(uint32_t id)
 {
 	ut_ad(!is_system_tablespace(id));
-	fil_space_t* space = fil_space_t::check_pending_operations(id);
+	fil_space_t* space = fil_space_t::drop(id, nullptr);
 	if (!space) {
 		return;
 	}
 
 	space->x_lock();
+	ut_ad(space->is_stopping());
 
 	/* Invalidate in the buffer pool all pages belonging to the
-	tablespace. Since we have invoked space->set_stopping(), readahead
+	tablespace. Since space->is_stopping() holds, readahead
 	can no longer read more pages of this tablespace to buf_pool.
 	Thus we can clean the tablespace out of buf_pool
 	completely and permanently. */
 	while (buf_flush_list_space(space));
-	ut_ad(space->is_stopping());
 
-	/* If it is a delete then also delete any generated files, otherwise
-	when we drop the database the remove directory will fail. */
-
-	if (char* cfg_name = fil_make_filepath(space->chain.start->name,
-					       fil_space_t::name_type{},
-					       CFG, false)) {
-		os_file_delete_if_exists(innodb_data_file_key, cfg_name, NULL);
-		ut_free(cfg_name);
-	}
-
-	/* If the free is successful, the wrlock will be released before
-	the space memory data structure is freed. */
-
-	if (!fil_space_free(id, true)) {
-		space->x_unlock();
+	space->x_unlock();
+	log_sys.latch.wr_lock(SRW_LOCK_CALL);
+	if (space->max_lsn != 0) {
+		ut_d(space->max_lsn = 0);
+		fil_system.named_spaces.remove(*space);
 	}
+	log_sys.latch.wr_unlock();
+	fil_space_free_low(space);
 }
 
 /** Delete a tablespace and associated .ibd file.
@@ -1628,19 +1811,8 @@ pfs_os_file_t fil_delete_tablespace(uint32_t id)
 {
   ut_ad(!is_system_tablespace(id));
   pfs_os_file_t handle= OS_FILE_CLOSED;
-  if (fil_space_t *space= fil_space_t::check_pending_operations(id))
-  {
-    /* Before deleting the file(s), persistently write a log record. */
-    mtr_t mtr;
-    mtr.start();
-    mtr.log_file_op(FILE_DELETE, id, space->chain.start->name);
-    handle= space->chain.start->handle;
-    mtr.commit_file(*space, nullptr);
-
+  if (fil_space_t *space= fil_space_t::drop(id, &handle))
     fil_space_free_low(space);
-  }
-
-  ibuf_delete_for_discarded_space(id);
   return handle;
 }
 
@@ -1854,12 +2026,16 @@ fil_ibd_create(
 
 	mtr.start();
 	mtr.log_file_op(FILE_CREATE, space_id, path);
-	mtr.commit();
-	log_write_up_to(mtr.commit_lsn(), true);
+	log_sys.latch.wr_lock(SRW_LOCK_CALL);
+	auto lsn= mtr.commit_files();
+	log_sys.latch.wr_unlock();
+	mtr.flag_wr_unlock();
+	log_write_up_to(lsn, true);
 
-	ulint type;
 	static_assert(((UNIV_ZIP_SIZE_MIN >> 1) << 3) == 4096,
 		      "compatibility");
+#if defined _WIN32 || defined HAVE_FCNTL_DIRECT
+	ulint type;
 	switch (FSP_FLAGS_GET_ZIP_SSIZE(flags)) {
 	case 1:
 	case 2:
@@ -1868,6 +2044,9 @@ fil_ibd_create(
 	default:
 		type = OS_DATA_FILE;
 	}
+#else
+	constexpr auto type = OS_DATA_FILE;
+#endif
 
 	file = os_file_create(
 		innodb_data_file_key, path,
@@ -1947,16 +2126,20 @@ err_exit:
 	DBUG_EXECUTE_IF("checkpoint_after_file_create",
 			log_make_checkpoint(););
 
+	mysql_mutex_lock(&fil_system.mutex);
 	if (fil_space_t* space = fil_space_t::create(space_id, flags,
 						     FIL_TYPE_TABLESPACE,
-						     crypt_data, mode)) {
+						     crypt_data, mode, true)) {
 		fil_node_t* node = space->add(path, file, size, false, true);
 		IF_WIN(node->find_metadata(), node->find_metadata(file, true));
+		mysql_mutex_unlock(&fil_system.mutex);
 		mtr.start();
 		mtr.set_named_space(space);
 		ut_a(fsp_header_init(space, size, &mtr) == DB_SUCCESS);
 		mtr.commit();
 		return space;
+	} else {
+		mysql_mutex_unlock(&fil_system.mutex);
 	}
 
 	if (space_name.data()) {
@@ -2071,6 +2254,10 @@ func_exit:
 		must_validate = true;
 	}
 
+	const bool operation_not_for_export =
+	  srv_operation != SRV_OPERATION_RESTORE_EXPORT
+	  && srv_operation != SRV_OPERATION_EXPORT_RESTORED;
+
 	/* Always look for a file at the default location. But don't log
 	an error if the tablespace is already open in remote or dict. */
 	ut_a(df_default.filepath());
@@ -2081,6 +2268,7 @@ func_exit:
 	drop_garbage_tables_after_restore() a little later. */
 
 	const bool strict = validate && !tablespaces_found
+		&& operation_not_for_export
 		&& !(srv_operation == SRV_OPERATION_NORMAL
 		     && srv_start_after_restore
 		     && srv_force_recovery < SRV_FORCE_NO_BACKGROUND
@@ -2129,7 +2317,9 @@ func_exit:
 			goto corrupted;
 		}
 
-		os_file_get_last_error(true);
+		if (!operation_not_for_export) {
+			goto corrupted;
+		}
 		sql_print_error("InnoDB: Could not find a valid tablespace"
 				" file for %.*s. %s",
 				static_cast<int>(name.size()), name.data(),
@@ -2207,8 +2397,10 @@ skip_validate:
 					    first_page)
 		: NULL;
 
+	mysql_mutex_lock(&fil_system.mutex);
 	space = fil_space_t::create(id, flags, purpose, crypt_data);
 	if (!space) {
+		mysql_mutex_unlock(&fil_system.mutex);
 		goto error;
 	}
 
@@ -2218,6 +2410,7 @@ skip_validate:
 	space->add(
 		df_remote.is_open() ? df_remote.filepath() :
 		df_default.filepath(), OS_FILE_CLOSED, 0, false, true);
+	mysql_mutex_unlock(&fil_system.mutex);
 
 	if (must_validate && !srv_read_only_mode) {
 		df_remote.close();
@@ -2296,6 +2489,7 @@ fil_ibd_discover(
 		case SRV_OPERATION_RESTORE:
 			break;
 		case SRV_OPERATION_NORMAL:
+		case SRV_OPERATION_EXPORT_RESTORED:
 			size_t len= strlen(db);
 			if (len <= 4 || strcmp(db + len - 4, dot_ext[IBD])) {
 				break;
@@ -2351,6 +2545,17 @@ fil_ibd_discover(
 	/* A datafile was not discovered for the filename given. */
 	return(false);
 }
+
+bool fil_crypt_check(fil_space_crypt_t *crypt_data, const char *f_name)
+{
+  if (crypt_data->is_key_found())
+    return true;
+  sql_print_error("InnoDB: Encryption key is not found for %s", f_name);
+  crypt_data->~fil_space_crypt_t();
+  ut_free(crypt_data);
+  return false;
+}
+
 /** Open an ibd tablespace and add it to the InnoDB data structures.
 This is similar to fil_ibd_open() except that it is used while processing
 the REDO log, so the data dictionary is not available and very little
@@ -2374,21 +2579,15 @@ fil_ibd_load(uint32_t space_id, const char *filename, fil_space_t *&space)
 	mysql_mutex_unlock(&fil_system.mutex);
 
 	if (space) {
-		/* Compare the filename we are trying to open with the
-		filename from the first node of the tablespace we opened
-		previously. Fail if it is different. */
-		fil_node_t* node = UT_LIST_GET_FIRST(space->chain);
-		if (0 != strcmp(innobase_basename(filename),
-				innobase_basename(node->name))) {
-			ib::info()
-				<< "Ignoring data file '" << filename
-				<< "' with space ID " << space->id
-				<< ". Another data file called " << node->name
-				<< " exists with the same space ID.";
-			space = NULL;
-			return(FIL_LOAD_ID_CHANGED);
-		}
-		return(FIL_LOAD_OK);
+		sql_print_information("InnoDB: Ignoring data file '%s'"
+				      " with space ID " ULINTPF
+				      ". Another data file called %s"
+				      " exists"
+				      " with the same space ID.",
+				      filename, space->id,
+				      UT_LIST_GET_FIRST(space->chain)->name);
+		space = NULL;
+		return FIL_LOAD_ID_CHANGED;
 	}
 
 	if (srv_operation == SRV_OPERATION_RESTORE) {
@@ -2496,16 +2695,17 @@ tablespace_check:
 					    first_page)
 		: NULL;
 
-	if (crypt_data && !crypt_data->is_key_found()) {
-		crypt_data->~fil_space_crypt_t();
-		ut_free(crypt_data);
+	if (crypt_data && !fil_crypt_check(crypt_data, filename)) {
 		return FIL_LOAD_INVALID;
 	}
 
+	mysql_mutex_lock(&fil_system.mutex);
+
 	space = fil_space_t::create(
 		space_id, flags, FIL_TYPE_TABLESPACE, crypt_data);
 
 	if (space == NULL) {
+		mysql_mutex_unlock(&fil_system.mutex);
 		return(FIL_LOAD_INVALID);
 	}
 
@@ -2517,6 +2717,7 @@ tablespace_check:
 	let fil_node_open() do that task. */
 
 	space->add(file.filepath(), OS_FILE_CLOSED, 0, false, false);
+	mysql_mutex_unlock(&fil_system.mutex);
 
 	return(FIL_LOAD_OK);
 }
@@ -2630,7 +2831,7 @@ inline void fil_node_t::complete_write()
   mysql_mutex_assert_not_owner(&fil_system.mutex);
 
   if (space->purpose != FIL_TYPE_TEMPORARY &&
-      srv_file_flush_method != SRV_O_DIRECT_NO_FSYNC &&
+      (!fil_system.is_write_through() && !my_disable_sync) &&
       space->set_needs_flush())
   {
     mysql_mutex_lock(&fil_system.mutex);
@@ -2654,8 +2855,8 @@ fil_io_t fil_space_t::io(const IORequest &type, os_offset_t offset, size_t len,
                          void *buf, buf_page_t *bpage)
 {
 	ut_ad(referenced());
-	ut_ad(offset % OS_FILE_LOG_BLOCK_SIZE == 0);
-	ut_ad((len % OS_FILE_LOG_BLOCK_SIZE) == 0);
+	ut_ad(offset % UNIV_ZIP_SIZE_MIN == 0);
+	ut_ad(len % 512 == 0); /* page_compressed */
 	ut_ad(fil_validate_skip());
 	ut_ad(type.is_read() || type.is_write());
 	ut_ad(type.type != IORequest::DBLWR_BATCH);
@@ -2734,10 +2935,6 @@ io_error:
 			     buf, offset, len);
 	}
 
-	/* We an try to recover the page from the double write buffer if
-	the decompression fails or the page is corrupt. */
-
-	ut_a(type.type == IORequest::DBLWR_RECOVER || err == DB_SUCCESS);
 	if (!type.is_async()) {
 		if (type.is_write()) {
 release_sync_write:
@@ -2757,67 +2954,64 @@ func_exit:
 
 #include <tpool.h>
 
-/** Callback for AIO completion */
-void fil_aio_callback(const IORequest &request)
+void IORequest::write_complete(int io_error) const
 {
   ut_ad(fil_validate_skip());
-  ut_ad(request.node);
+  ut_ad(node);
+  ut_ad(is_write());
+  node->complete_write();
 
-  if (!request.bpage)
+  if (!bpage)
   {
     ut_ad(!srv_read_only_mode);
-    if (request.type == IORequest::DBLWR_BATCH)
-      buf_dblwr.flush_buffered_writes_completed(request);
+    if (type == IORequest::DBLWR_BATCH)
+      buf_dblwr.flush_buffered_writes_completed(*this);
     else
-      ut_ad(request.type == IORequest::WRITE_ASYNC);
-write_completed:
-    request.node->complete_write();
-  }
-  else if (request.is_write())
-  {
-    buf_page_write_complete(request);
-    goto write_completed;
+      ut_ad(type == IORequest::WRITE_ASYNC);
   }
   else
+    buf_page_write_complete(*this, io_error);
+
+  node->space->release();
+}
+
+void IORequest::read_complete(int io_error) const
+{
+  ut_ad(fil_validate_skip());
+  ut_ad(node);
+  ut_ad(is_read());
+  ut_ad(bpage);
+
+  const page_id_t id(bpage->id());
+
+  if (UNIV_UNLIKELY(io_error != 0))
   {
-    ut_ad(request.is_read());
-
-    /* IMPORTANT: since i/o handling for reads will read also the insert
-    buffer in fil_system.sys_space, we have to be very careful not to
-    introduce deadlocks. We never close fil_system.sys_space data
-    files and never issue asynchronous reads of change buffer pages. */
-    const page_id_t id(request.bpage->id());
-
-    if (dberr_t err= request.bpage->read_complete(*request.node))
+    sql_print_error("InnoDB: Read error %d of page " UINT32PF " in file %s",
+                    io_error, id.page_no(), node->name);
+    buf_pool.corrupted_evict(bpage, buf_page_t::READ_FIX);
+  corrupted:
+    if (recv_recovery_is_on() && !srv_force_recovery)
     {
-      if (recv_recovery_is_on() && !srv_force_recovery)
-      {
-        mysql_mutex_lock(&recv_sys.mutex);
-        recv_sys.set_corrupt_fs();
-        mysql_mutex_unlock(&recv_sys.mutex);
-      }
-
-      if (err != DB_FAIL)
-        ib::error() << "Failed to read page " << id.page_no()
-                    << " from file '" << request.node->name << "': " << err;
+      mysql_mutex_lock(&recv_sys.mutex);
+      recv_sys.set_corrupt_fs();
+      mysql_mutex_unlock(&recv_sys.mutex);
     }
   }
+  else if (dberr_t err= bpage->read_complete(*node))
+  {
+    if (err != DB_FAIL)
+      ib::error() << "Failed to read page " << id.page_no()
+                  << " from file '" << node->name << "': " << err;
+    goto corrupted;
+  }
 
-  request.node->space->release();
+  node->space->release();
 }
 
 /** Flush to disk the writes in file spaces of the given type
 possibly cached by the OS. */
 void fil_flush_file_spaces()
 {
-  if (srv_file_flush_method == SRV_O_DIRECT_NO_FSYNC)
-  {
-    ut_d(mysql_mutex_lock(&fil_system.mutex));
-    ut_ad(fil_system.unflushed_spaces.empty());
-    ut_d(mysql_mutex_unlock(&fil_system.mutex));
-    return;
-  }
-
 rescan:
   mysql_mutex_lock(&fil_system.mutex);
 
@@ -2946,27 +3140,12 @@ fil_space_validate_for_mtr_commit(
 	ut_ad(!is_predefined_tablespace(space->id));
 
 	/* We are serving mtr_commit(). While there is an active
-	mini-transaction, we should have !space->stop_new_ops. This is
+	mini-transaction, we should have !space->is_stopping(). This is
 	guaranteed by meta-data locks or transactional locks. */
-	ut_ad(!space->is_stopping()
-	      || space->is_being_truncated /* fil_truncate_prepare() */
-	      || space->referenced());
+	ut_ad(!space->is_stopping() || space->referenced());
 }
 #endif /* UNIV_DEBUG */
 
-/** Write a FILE_MODIFY record for a persistent tablespace.
-@param[in]	space	tablespace
-@param[in,out]	mtr	mini-transaction */
-static
-void
-fil_names_write(
-	const fil_space_t*	space,
-	mtr_t*			mtr)
-{
-	ut_ad(UT_LIST_GET_LEN(space->chain) == 1);
-	fil_name_write(space->id, UT_LIST_GET_FIRST(space->chain)->name, mtr);
-}
-
 /** Note that a non-predefined persistent tablespace has been modified
 by redo log.
 @param[in,out]	space	tablespace */
@@ -2974,7 +3153,9 @@ void
 fil_names_dirty(
 	fil_space_t*	space)
 {
-	mysql_mutex_assert_owner(&log_sys.mutex);
+#ifndef SUX_LOCK_GENERIC
+	ut_ad(log_sys.latch.is_write_locked());
+#endif
 	ut_ad(recv_recovery_is_on());
 	ut_ad(log_sys.get_lsn() != 0);
 	ut_ad(space->max_lsn == 0);
@@ -2984,56 +3165,48 @@ fil_names_dirty(
 	space->max_lsn = log_sys.get_lsn();
 }
 
-/** Write FILE_MODIFY records when a non-predefined persistent
-tablespace was modified for the first time since the latest
-fil_names_clear().
-@param[in,out]	space	tablespace */
-void fil_names_dirty_and_write(fil_space_t* space)
+/** Write a FILE_MODIFY record when a non-predefined persistent
+tablespace was modified for the first time since fil_names_clear(). */
+ATTRIBUTE_NOINLINE ATTRIBUTE_COLD void mtr_t::name_write()
 {
-	mysql_mutex_assert_owner(&log_sys.mutex);
-	ut_d(fil_space_validate_for_mtr_commit(space));
-	ut_ad(space->max_lsn == log_sys.get_lsn());
+#ifndef SUX_LOCK_GENERIC
+  ut_ad(log_sys.latch.is_write_locked());
+#endif
+  ut_d(fil_space_validate_for_mtr_commit(m_user_space));
+  ut_ad(!m_user_space->max_lsn);
+  m_user_space->max_lsn= log_sys.get_lsn();
 
-	fil_system.named_spaces.push_back(*space);
-	mtr_t mtr;
-	mtr.start();
-	fil_names_write(space, &mtr);
+  fil_system.named_spaces.push_back(*m_user_space);
+  ut_ad(UT_LIST_GET_LEN(m_user_space->chain) == 1);
 
-	DBUG_EXECUTE_IF("fil_names_write_bogus",
-			{
-				char bogus_name[] = "./test/bogus file.ibd";
-				fil_name_write(
-					SRV_SPACE_ID_UPPER_BOUND,
-					bogus_name, &mtr);
-			});
-
-	mtr.commit_files();
+  mtr_t mtr;
+  mtr.start();
+  fil_name_write(m_user_space->id,
+                 UT_LIST_GET_FIRST(m_user_space->chain)->name,
+                 &mtr);
+  mtr.commit_files();
 }
 
 /** On a log checkpoint, reset fil_names_dirty_and_write() flags
-and write out FILE_MODIFY and FILE_CHECKPOINT if needed.
-@param[in]	lsn		checkpoint LSN
-@param[in]	do_write	whether to always write FILE_CHECKPOINT
-@return whether anything was written to the redo log
-@retval false	if no flags were set and nothing written
-@retval true	if anything was written to the redo log */
-bool
-fil_names_clear(
-	lsn_t	lsn,
-	bool	do_write)
+and write out FILE_MODIFY if needed, and write FILE_CHECKPOINT.
+@param lsn  checkpoint LSN
+@return current LSN */
+ATTRIBUTE_COLD lsn_t fil_names_clear(lsn_t lsn)
 {
 	mtr_t	mtr;
 
-	mysql_mutex_assert_owner(&log_sys.mutex);
+#ifndef SUX_LOCK_GENERIC
+	ut_ad(log_sys.latch.is_write_locked());
+#endif
 	ut_ad(lsn);
+	ut_ad(log_sys.is_latest());
 
 	mtr.start();
 
 	for (auto it = fil_system.named_spaces.begin();
 	     it != fil_system.named_spaces.end(); ) {
-		if (mtr.get_log_size()
-		    + strlen(it->chain.start->name)
-		    >= RECV_SCAN_SIZE - (3 + 5 + 1)) {
+		if (mtr.get_log_size() + strlen(it->chain.start->name)
+		    >= recv_sys.MTR_SIZE_MAX - (3 + 5)) {
 			/* Prevent log parse buffer overflow */
 			mtr.commit_files();
 			mtr.start();
@@ -3056,20 +3229,13 @@ fil_names_clear(
 		was called. If we kept track of "min_lsn" (the first LSN
 		where max_lsn turned nonzero), we could avoid the
 		fil_names_write() call if min_lsn > lsn. */
-
-		fil_names_write(&*it, &mtr);
-		do_write = true;
-
+		ut_ad(UT_LIST_GET_LEN((*it).chain) == 1);
+		fil_name_write((*it).id, UT_LIST_GET_FIRST((*it).chain)->name,
+			       &mtr);
 		it = next;
 	}
 
-	if (do_write) {
-		mtr.commit_files(lsn);
-	} else {
-		ut_ad(!mtr.has_modifications());
-	}
-
-	return(do_write);
+	return mtr.commit_files(lsn);
 }
 
 /* Unit Tests */
